/* =================================================================
 *
 * Copyright (C) 1989, 1990, 1993 Linas Vepstas
 *
 *                   All Rights Reserved
 *
 * Permission to use, copy, and modify this software and its 
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies 
 * and that both the copyright notice and this permission notice 
 * appear in supporting documentation.
 *
 * Permission to redistribute this software and its documentation
 * is hereby granted, provided that such redistribution is not 
 * performed for monetary profit or gain.
 *
 * Redistribution for profit without express written consent is
 * prohibited.  Contact the author if you wish to use this software 
 * and/or its documentation in a commercial product.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND.  
 * THE AUTHOR DISCLAIMS ALL EXPRESS AND IMPLIED WARRANTIES, INCLUDING 
 * BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * IN NO EVENT SHALL THE AUTHOR BE HELD LIABLE FOR ANY SPECIAL, DIRECT, 
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING 
 * IN LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, 
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION 
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Bug reports or fixes may be sent to the author, who may or may not 
 * act upon them as he desires.
 * 
 * The author's last known address is:
 * 
 *                       Linas Vepstas
 *                     1518 Enfield Road
 *                  Austin, Texas 78703-3424
 *                      1-(512)-499-8246
 *                    linas@austin.ibm.com
 *
 * =================================================================
 */

/* Linas Vepstas -- May 1989 
 */


#include <sys/signal.h>         /* required for ? */
#include <string.h>		/* required for string ops (on prog names) */
#include "may.h"
#include "maymsg.h"


/* maintain a list of all open communications channels */
struct mayProcId *open_socks = NULL;

struct mayProcId *The_Master_Sock = NULL;

int	listener_sockfd; 	/* daemon's "listener" socket */

char 	service[] = "maydaemon";   /* used ot get service port number */

void 	maySpawnUdp ();
void 	maySpawnTcp ();

/* ========================================================= */
/* 
 * The maydaemon performs several functions:
 * 
 * *) one is to intercept "maycreate" messages.  Upon receiving such a
 * message, it will fork and execute the program named in the "maycreate"
 * message (the "child").  Before starting the execution of the child, a
 * socket is opened, which will be inherited by the child. A reply
 * confirming sucessful start of the child is returned.  The reply contains
 * the address of the child's socket.
 */

#ifdef ANSI_C
main (int argc, char *argv[])
#else 
main(argc,argv)
int argc;
char *argv[];
#endif
{

   struct mayProcId *tmp;
   struct	mayMsghead		header;
   int 		message_len;
   char 	read_buf [2048];	/* read buffer */
   int		bufflen, flag;
   long		loop_counter;

   fprintf (stderr, " %s \n", mayPatchLevel);

#ifdef ANSI_C
   fprintf (stderr, "Compiled at %s on %s \n", __TIME__, __DATE__);
#endif 

   /* initialize -- i.e. open a TCP socket. */
   maydaemon_init ();

   /* 
    * Do a UDP initialization also.  The UDP socket is needed so that we
    * can send stuff (specifically, the parents address) to the spawned
    * UDP child (that is, a spawned child that has a UDP connection).
    */
   tmp = mayInitUdp();
   if (tmp != NULL) {
      /* returned values are not needed */
      if (tmp->hostname != NULL) free (tmp->hostname);
      free (tmp);
   }

   loop_counter = 0;

   /* 
    * This flag enables alternate behaviour of some of the MAY routines
    * so that they are apporpriate for the daemon 
    */
   may_internal_flag = MAY_INTERNAL_FLAG; 

   /* here, we begin looping, waiting for clients to ask for a
    * connection.  Presumably, once the connection has been made, a message
    * will come across.  
    * These messages will be "daemon private" messages, instructing the
    * may daemon to perform various tasks.
    * After processing precisely one message, the connection is shut down.
    */


   while (TRUE) {
      loop_counter ++;

      /* Try to accept an incomming connection.
       * Note that the accept call blocks until a connection comes in.
       */
      if (0>(The_Master_Sock->socket_id = accept(listener_sockfd,0,0))) {
         perror ("maydaemon error: accept failed");
         continue;
      }

#ifdef DEBUG
      printf ("maydaemon just accepted connection ");
      printf ("listen fd = %d accept fd = %d \n", listener_sockfd,
         The_Master_Sock->socket_id);
#endif

      header.msgptr = (Void *) read_buf;

      /* get any and all messages
       * Use a short timeout so that the daemon isn't tied up a long
       * time if other end fails to transmit.  Maybe the daemon should 
       * fork at this point to make sure it doesn't croak if the 
       * network screws up ?
       */
      bufflen = 2048;
      flag = mayReceive (NULL, NULL, NULL, 10, &header, &bufflen);
      if (flag == MAY_TIMED_OUT) { 
         perror("maydaemon error: other end never sent anything");
         close (The_Master_Sock->socket_id);
         continue;
         }

      if ((flag == MAY_CLOSE_CONNECTION) || (flag == MAY_NO_OPEN_SOCKS)) { 
         perror("maydaemon error: connection was dropped");
         continue;
         }

      if ( flag < 0 ) {
         perror("maydaemon error: mayrecv");
         close (The_Master_Sock->socket_id);
         continue;
         }

      printf ("daemon connect %d \n", loop_counter);

      /* if the message is not for us, throw it away */
      if (header.msgtype != DAEMONMSG) continue;

      /* parse according to message subtype */
      if (header.subtype == MAY_CREATE_TCP) maySpawnTcp (&header);
      if (header.subtype == MAY_CREATE_UDP) maySpawnUdp (&header);

      /* now that we sent the reply, close down the socket */
      if (0 != close (The_Master_Sock->socket_id)) {
         perror("maydaemon error: unable to close socket");
         continue;
      }
   }
}


/* ========================================================= */
/* 
 * maydaemon_init
 * Performs general initalization for the daemon
 *
 * maydaemon_init pseudocode:
 * ==> on power up, maydaemon finds out about its host
 * ==> it sets up the socket on which it will receive.  (The Master Sock)
 * ==> sets the SIGCLD signal to SIG_IGN (avoid zombies)
 */

maydaemon_init () {

   struct 	hostent	*local_hostent;	/* local host addresses */
   struct	servent	*local_servent;	/* local service port number */
   char 	local_host_name [512];	/* local host name */

   /* get the name of the local host */
   if (0 != gethostname (local_host_name,  sizeof (local_host_name)))
      perror ("maydaemon error: gethostname");

   /* get local host adresses */
   if((local_hostent = gethostbyname (local_host_name)) == NULL) {
      perror ("maydaemon error: gethostbyname");
      exit (EXIT_FAILURE);
   }

   /* get the maydaemon service port number */
   if((local_servent = getservbyname (service,0)) == NULL) {
      perror ("maydaemon error: getservbyname");
      exit (EXIT_FAILURE);
   }

   /* set up a connection list so that mayReceive can do its stuff */
   The_Master_Sock = (struct mayProcId *) malloc (SIZEOPRO);
   if( NULL == The_Master_Sock ) {
      perror ("maydaemon error: unable to malloc");
      exit (EXIT_FAILURE);
   }
   memset (The_Master_Sock, 0, SIZEOPRO); /* you WILL bomb if you don't do this */

   /* open a tcp/ip socket */
   The_Master_Sock->proto = SOCK_STREAM;
   if((The_Master_Sock->socket_id = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
      perror ("maydaemon error: unable to open stream socket");
      exit (EXIT_FAILURE);
   }

   /* construct a name for the socket */
   The_Master_Sock->address.sin_family = AF_INET;
   The_Master_Sock->address.sin_port = htons (local_servent->s_port);
   memcpy (&The_Master_Sock->address.sin_addr.s_addr, 
      local_hostent->h_addr, local_hostent->h_length);

#ifdef DEBUG
   printf (" maydaemon has opened stream socket ");
   printf ( " address %d 0x%x %d \n",
   The_Master_Sock->address.sin_family,
   ntohs (The_Master_Sock->address.sin_addr.s_addr),
   The_Master_Sock->address.sin_port);
#endif

   /* bind the name to the socket */
   if (0 != bind (The_Master_Sock->socket_id, 
                 &The_Master_Sock->address, SIZEOSOX)) {
      perror ("maydaemon error: unable to bind to stream socket");
      exit (EXIT_FAILURE);
   }

   /* listen for a connection, ready to accept a backlog of 10
    * connections */
   if (0 != listen (The_Master_Sock->socket_id, 10)) {
      perror ("maydaemon error: unable to listen to stream socket");
      exit (EXIT_FAILURE);
   }

#ifdef DEBUG
   printf (" maydaemon listening on sockfd %d ", The_Master_Sock->socket_id);
   printf ( " address %d 0x%x %d \n",
   The_Master_Sock->address.sin_family,
   ntohs (The_Master_Sock->address.sin_addr.s_addr),
   The_Master_Sock->address.sin_port);
#endif

   /* Disable the "death of child" signal, so that a bunch of
    * zombies don't go clogging up the process table */
   if (SIG_ERR == signal (SIGCLD, SIG_IGN)) {
      perror ("maydaemon error: signal");
      exit (EXIT_FAILURE);
   }
   
   /* This should nice me down to 31 so just about everything runs at
    * higher priority */
   if (-1 == nice (11)) {
      perror ("maydaemon error: unable to nice down priority");
      exit (EXIT_FAILURE);
   }

   /* the file descriptor we've created will be the listener fd, save it. */
   listener_sockfd = The_Master_Sock->socket_id;

   /* chain it into the list of sockets, so that we can use mayRecv on this sock */
   open_socks =  The_Master_Sock;
   open_socks -> next = NULL;
}



/* ========================================================= */
/* 
 * maySpawnTcp
 * Spawns a process with a TCP connection back to the parent.
 * 
 * This routine starts up the requested process. It does so by fork and
 * execv.  It does not need to open any new sockets -- it simply inherits
 * the stream socket from the daemon.
 * 
 * It will fork and execute the program named in the "maycreate"
 * message (the "child").  A reply confirming sucessful start of 
 * the child is returned.  
 * 
 */

#ifdef ANSI_C
void maySpawnTcp (struct mayMsghead *head)
#else
void maySpawnTcp (head)
struct	mayMsghead 	*head;
#endif
{

   struct 	create_reply 	cr; 	/* create reply message */
   int		pid;			/* process id after fork */

   /* if child sock create fails, the parent will be
   informed.  But for now, start with a clean slate */
   cr.errortype = 0;
   cr.err_msg[0] = NULL;

   /* fork */
   pid = fork ();
   if (pid < 0) {
      perror ("maydaemon error: maySpawnTcp: unable to fork");
      cr.errortype = 4;
      strcpy (cr.err_msg, "maydaemon error: maySpawnTcp: Unable to fork");
      if (0 > maySend (&head->sproc, DAEMONMSG, CREATE_REPLY, 
         (Void *) &cr, sizeof (struct create_reply))) {
         perror ("maydaemon: maySpawnTcp: error: maysend create reply");
      }
   }

   else if (pid == 0) {
      /* if we are here, we are the child */

      /* We need to tell the child process what its socket connections
       * are going to be.  Open a file, write the pertinent info into 
       * it. This is necessary since some operating systems seem unable
       * to count correctly, and therefore the child can't "guess" what 
       * the correct socket id's will be.
       */
      mayDumpId (head->sproc.socket_id);

      /* send the reply message back to whomever sent us the create msg */
      if (0 > maySend (&head->sproc, DAEMONMSG, CREATE_REPLY, 
         (Void *) &cr, sizeof (struct create_reply))) {
         perror ("maydaemon: maySpawnTcp: error: maysend create reply");
      }

      /* close the daemon's "listener" socket, which, incidentally,
       * should be socket number 3. Also, close the udp socket, which, 
       * incidentally, should be socket number 4. */
      close (listener_sockfd);
      close (The_udp_sock->socket_id);

#ifdef DEBUG
      printf ("maySpawnTcp about to execv %s \n", 
          ((struct create_msg *) head->msgptr) -> prog_name);
#endif

      /* overlay the daemon with the child */
      mayExec ( ((struct create_msg *) head->msgptr) -> prog_name);
      /* if we ever get here, the overlay failed */
      perror ("maydaemon: maySpawnTcp: error: execv");
      exit (EXIT_FAILURE);
   }

   else {
      /* if we are here, we are still the daemon */
   }
}

/* ========================================================= */
/* 
 * maySpawnUdp
 * Spawns a process with a UDP mailing address
 * 
 * This is the routine that opens up a socket for the child, forks, sends a
 * confirm message back to parent, sends the parents address to child.
 * 
 * It will fork and execute the program named in the "maycreate"
 * message (the "child").  Before starting the execution of the child, a
 * socket is opened, which will be inherited by the child. A reply
 * confirming sucessful start of the child is returned.  The reply contains
 * the address of the childs socket.
 * 
 */

#ifdef ANSI_C
void maySpawnUdp (struct mayMsghead *head)
#else 
void maySpawnUdp (head)
struct	mayMsghead 	*head;
#endif
{

   struct	mayProcId	child_proc;
   struct	mayProcId	*tmp;
   int		name_len;		/* should equal SIZEOFSOX */
   struct 	create_reply 	cr; 	/* create reply message */
   int		pid;			/* process id after fork */
   int 		message_len;

   /* if child sock create fails, the parent will be
   informed.  But for now, start with a clean slate */
   cr.errortype = 0;
   cr.err_msg[0] = NULL;

   /* open the socket which the child will inherit */
   child_proc.proto = SOCK_DGRAM;
   child_proc.socket_id = socket (AF_INET, SOCK_DGRAM, 0);

#ifdef DEBUG
   printf ("maySpawnUDP opened sockfd = %d \n", child_proc.socket_id);
#endif

   /* turn this into a non-blocking socket */
   /* this code is obsolete -- but may still be useful in porting to old
    * operating systems. */
   /* if (-1 == ioctl (child_sock_id, FIONBIO, junk)) {
      perror (" maydaemon error: ioctl (child)");
      cr.errortype = 1;
      strcpy (cr.err_msg, "maydaemon: Unable to unblock socket for child");
      }  */

   /* construct the socket name -- same host addr, differnt port */
   memcpy (&child_proc.address, &The_udp_sock->address, SIZEOSOX);
   child_proc.address.sin_port = 0; /* port automatically assigned */

   /* bind the name to the socket */
   if (0 != bind (child_proc.socket_id, &child_proc.address, SIZEOSOX)) {
      perror("maydaemon: maySpawnUdp: error on bind (child)");
      cr.errortype = 2;
      strcpy (cr.err_msg, "remote maydaemon: Unable to bind socket for child");
   }

   /* find out the port which was assigned to the child */
   name_len = SIZEOSOX;
   if (0 != getsockname (child_proc.socket_id, &child_proc.address, &name_len)){
      perror ("maydaemon error: getsockname");
      cr.errortype = 3;
      strcpy (cr.err_msg, "remote maydaemon: Unable to get child socket name");
   }

   /* place the childs address into the create reply message */
   memcpy (&cr.sproc, &child_proc, SIZEOPRO);

#ifdef DEBUG
   printf(" create about to send reply to fd:%d addr: %d 0x%x %d \n",
   head->sproc.socket_id,
   head->sproc.address.sin_family, 
   head->sproc.address.sin_addr.s_addr, 
   head->sproc.address.sin_port );
#endif

   /* send the reply message back to whomever sent us the create msg */
   if (0 > maySend (&head->sproc, DAEMONMSG, CREATE_REPLY, (Void *) &cr,
      sizeof (struct create_reply))) {
      perror ("maydaemon: maySpwanUdp: error on maysend (create reply)");
   }

   /* fork */
   pid = fork ();
   if (pid < 0) {
      perror ("maydaemon: maySpawnUdp: error on fork");
   }
   else if (pid == 0) {
      /* if we are here, we are the child */

      /* Lets see -- file descriptor 3 is daemon's tcp "listener" socket,
       *                             4 is daemon's udp sendto socket,
       *                             5 is daemon's actual tcp connection
       * therefore, we better be 6. However, some operating systems cannot
       * count sequentially, so therefore,
       * we need to have a way of communicating to the child what it's 
       * socket id will be.  Do this by writing it into a file */
      mayDumpId (child_proc.socket_id);

      /* close the daemons tcp "listener" socket (# 3), udp "sendto"
       * sock (# 4) and tcp active sock (# 5) */
      close (listener_sockfd);
      close (The_udp_sock->socket_id);
      close (The_Master_Sock->socket_id);
      
      /* overlay the daemon with the child */
      mayExec( ((struct create_msg *) head->msgptr) -> prog_name);

      /* if we ever get here, the overlay failed */
      perror ("maydaemon: maySpawnUdp: error on execv");
      exit (EXIT_FAILURE);
   }
   else {
      /* if we are here, we are still the daemon */

      /* close the childs socket. */
      /* do only a close. Do Not even THINK of doing a shutdown */
      close (child_proc.socket_id);

      /* I know this looks ugly but ... we send on the DEAMON'S sockfd,
       * TO the CHILDS address */
      child_proc.socket_id = The_udp_sock->socket_id;

      /* send the address of parent to the child */
      tmp = &(((struct create_msg *) head->msgptr) -> sproc);
#ifdef DEBUG
      printf(" maySpawnUdp sends this as parent addr: %d 0x%x %d \n",
                                   tmp -> address.sin_family, 
                                   tmp -> address.sin_addr.s_addr, 
                                   tmp -> address.sin_port );
#endif

      if (0 > maySend (&child_proc, DAEMONMSG, PARENTADDR,
         (Void *) tmp, SIZEOPRO)) {
         perror ("maydaemon: maySpawnUdp: error on maysend (addr to child)");
      }
   }
}

/* ========================================================= */
/* 
 * mayDumpId 
 * Writes file descriptor to temporary file
 *
 * The file written to is uniquely identified by the process pid.
 */

void mayDumpId (sid)
int sid;
{
   int 		pid;			/* process id */
   char		filename[30];		/* file containing socket id's */
   FILE		*fid;			/* file id of above */

   pid = getpid();
   sprintf (filename, "/tmp/may%d", pid);

#ifdef DEBUG
   printf ("/tmp/may%d\n", pid);
#endif

   fid = fopen (filename, "w");

   if (fid == NULL) {
      perror ("maydaemon: mayDumpId: error: unable to open file");
      exit (EXIT_FAILURE);
   }

   fprintf (fid, "%d", sid);
   fclose (fid);
}

/* ========================================================= */
/* 
 * mayExec 
 * exec a program
 *
 * This routine performs (for all practical purposes) an execv.
 * The value add that it provides is that it parses the argumnet
 * string for arguments.  (That is, it accepts a string which is the
 * program name followed by arguments, and it splits out the arguments
 * into a form suitable for execv).
 */

#ifdef ANSI_C
void mayExec (char *pname)
#else
void mayExec (pname)
char *pname;
#endif
{
   char **argv;
   char *chptr;
   int i;

   /* malloc enough for 100 arguments */
   argv = (char**) malloc (100 * sizeof(char *));

   /* the first argument must be the program name */
   if (NULL == (chptr = strtok (pname, "\t\n ,;"))) {
      perror(" mayExec error: no path specified ");
   }

   argv[0] = (char *) malloc (strlen(chptr)+1);
   strcpy (argv[0], chptr);

   i=1;
   /* get the rest of the command line arguments */
   while (NULL != (chptr = strtok (NULL, "\t\n ,;"))) {
      argv[i] = (char *) malloc (strlen(chptr)+1);
      strcpy (argv[i], chptr);
      i++;
   }
   argv[i] = NULL;

   /* overlay the daemon with the child */
   execv ( argv[0], argv);

}

/* ======================== end of file ======================= */
