/* =================================================================
 *
 * Copyright (C) 1989, 1990, 1993 Linas Vepstas
 *
 *                   All Rights Reserved
 *
 * Permission to use, copy, and modify this software and its 
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies 
 * and that both the copyright notice and this permission notice 
 * appear in supporting documentation.
 *
 * Permission to redistribute this software and its documentation
 * is hereby granted, provided that such redistribution is not 
 * performed for monetary profit or gain.
 *
 * Redistribution for profit without express written consent is
 * prohibited.  Contact the author if you wish to use this software 
 * and/or its documentation in a commercial product.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND.  
 * THE AUTHOR DISCLAIMS ALL EXPRESS AND IMPLIED WARRANTIES, INCLUDING 
 * BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * IN NO EVENT SHALL THE AUTHOR BE HELD LIABLE FOR ANY SPECIAL, DIRECT, 
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING 
 * IN LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, 
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION 
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Bug reports or fixes may be sent to the author, who may or may not 
 * act upon them as he desires.
 * 
 * The author's last known address is:
 * 
 *                       Linas Vepstas
 *                     1518 Enfield Road
 *                  Austin, Texas 78703-3424
 *                      1-(512)-499-8246
 *                    linas@austin.ibm.com
 *
 * =================================================================
 */

#include <stdio.h>	/* required for file operations */
#include "may.h"
#include "maymsg.h"

char *mayPatchLevel = "MAY Distributed Processing Package Version 2.09";

#ifdef IBM_AIXv221
#define		remove(file)	unlink(file)
#endif

extern struct hostent *gethostbyname();

struct mayProcId *The_udp_sock = NULL;

/* ============================================================== */
/* mayInitUdp 
 * Initializes a process for using a UDP connection
 *
 * mayInitUdp is an initialization routine.  
 * It will create and initialize a UDP connectionless socket.
 * It returns a UDP socket address for the calling routine
 *  (i.e. the address to which others can send messages.)
 *
 * If the mayInit call is made in the parent,
 * *) a new socket is opened, initialized, etc. and returned.
 * 
 * If the mayInit call is made by a "child" or spawned mayProcId,
 * *) the socket has already been set up by the local daemon.  All
 * that needs be done is to get the socket name
 */

#ifdef ANSI_C
struct mayProcId * mayInitUdp (void)
#else
struct mayProcId * mayInitUdp ()
#endif
{
   
   struct       mayProcId       *os;	/* kept internally */
   struct	mayProcId	*proc_ptr; /* returned to user */
   struct 	hostent 	*he; 	/* the hostent structure */
   int 		len;
   char 	host_name [512];		/* host name */

   /* allocate space for the list of known connections */
   os = (struct mayProcId *) malloc (SIZEOPRO);
   if (os == NULL) {
      perror ("mayInitUdp: error on malloc");
      return(NULL);
   }
   memset (os, 0, SIZEOPRO);  /* you WILL bomb if you don't do this */
   os->next = open_socks;
   open_socks = os;

   gethostname (host_name, 512);

   /* The following if test is an attempt to distinguish between 
    * a parent and a child process.  A different form of initialization takes
    * place depending on whether it is the parent of the child.  A parent
    * needs to create a socket, and do all that other garbage.  A child will
    * already have had all that garbage done for it by the deamon.  The daemon
    * forks, and the child inherits the socket. 
    *
    * We detect a child by looking to see if there is a temporary file
    * continaing its UDP socket descriptor present.  If present, then 
    * child, else parent.
    */

   mayGetId (&os->socket_id);
   if (os->socket_id == 0) {
      /* if the file wasn't found, we must be the parent */
      /* open a new socket */
      os->socket_id = socket (AF_INET, SOCK_DGRAM, 0);
      if (os->socket_id < 0) {
         perror ("mayInitUdp error: could not open socket");
         open_socks = open_socks->next;
         free (os);
         return(NULL);
      }

      /* make this a non-blocking socket; but first get devinfo */
      /* The AIX manuals say that I should use ioctl to unblock.
         However, ioctl doesn't always work (although it did work
         sometimes ...). The BSD manuals say I
         should use fcntl to unblock.  This does seem to always work.
         so... the following commented out code is the bad ioctl stuff
XXX      if (-1 == ioctl (may_sockfd, IOCINFO, &sock_device_info)) {
         perror ("mayInit error: ioctl IOCINFO");
         } 
      if (0 > ioctl (may_sockfd, FIONBIO, junk)) {
         perror ("mayInit error: ioctl FIONBIO");
         }  */
/*      if (0 > fcntl (may_sockfd, F_SETFL, O_NDELAY)) {
         perror ("mayInit error: fcntl F_SETFL O_NDELAY");
         } */


      /*begin all the steps necessary to construct socket address */
      he = gethostbyname (host_name);
      if (he == NULL) {
         perror ("mayInitUdp error: could not get local host name");
         open_socks = open_socks->next;
         free (os);
         return(NULL);
      }

      /* construct the socket address */
      os->address.sin_family = AF_INET;
      memcpy(&os->address.sin_addr.s_addr, he->h_addr, he->h_length);
      os->address.sin_port = 0;
      if (bind (os->socket_id, &os->address, SIZEOSOX) < 0) {
         perror ("mayInitUdp error: unable to bind socket");
         close(os->socket_id);
         open_socks = open_socks->next;
         free (os);
         return(NULL);
      }
   }

   else {
#ifdef DEBUG
      printf("mayinitUDP: think I'm the child \n");
#endif
   }

   /* get the port number */
   len = SIZEOSOX;
   if (0 > getsockname (os->socket_id, &os->address, &len)) {
      perror ("mayInitUdp error: unable to get socket name");
      close(os->socket_id);
      open_socks = open_socks->next;
      free (os);
      return(NULL);
   }
   os->proto = SOCK_DGRAM;

   /* make a record of the ASCII hostname */
   os->hostname = (char *) malloc (strlen(host_name)+1);
   if (os->hostname == NULL) {
      perror ("mayInitUdp error: could not malloc");
      close(os->socket_id);
      open_socks = open_socks->next;
      free (os);
      return(NULL);
   }
   strcpy (os->hostname, host_name);

   /* we need to keep track of this address for the "create" routine */
   The_udp_sock = os;

   /* OK-- we've got a socket.  Now make a copy of it for the user */
   /* allocate space for the thing that will be returned */
   proc_ptr = (struct mayProcId *) malloc (SIZEOPRO);
   if (proc_ptr == NULL) {
      perror ("mayInitUdp error: could not malloc");
      close (os->socket_id);
      open_socks = open_socks->next;
      free (os->hostname);
      free (os);
      return(NULL);
   }

   /* copy the process id */
   memcpy (proc_ptr, os, SIZEOPRO);
   proc_ptr->unread_bytes_remaining = 0;
   proc_ptr->partial_msg = NULL;

   proc_ptr->hostname = (char *) malloc (strlen(host_name)+1);
   if (proc_ptr->hostname == NULL) {
      perror ("mayInitUdp error: could not malloc");
      close(os->socket_id);
      open_socks = open_socks->next;
      free (os->hostname);
      free (os);
      return(NULL);
   }
   strcpy (proc_ptr->hostname, host_name);

#ifdef DEBUG
   printf ( " mayInitUdp constructed for itself fd %d addr %d 0x%x %d \n", 
   proc_ptr -> socket_id,
   proc_ptr -> address.sin_family,
   proc_ptr -> address.sin_addr.s_addr, 
   proc_ptr -> address.sin_port);
#endif

   /* return the malloced area */
   return (proc_ptr);

}

/* ============================================================== */
/* mayGetParentUdp 
 * Returns the address of the parent.
 * 
 * mayGetParentUdp is a call that is valid only when made by the child.
 * It looks at the childs socket, looks for a message from the maydaemon,
 * reads the message (which should contain the parent's address) and
 * returns that address.
 */

#ifdef ANSI_C
struct mayProcId * mayGetParentUdp (void)
#else
struct mayProcId * mayGetParentUdp ()
#endif
{
   
   struct 	mayProcId	*parent_proc_ptr;
   struct	mayMsghead		addr_msg;
   int 		flag;
   int 		len;

   /* allocate space for the thing that will be returned */
   parent_proc_ptr = (struct mayProcId *) malloc (SIZEOPRO);
   if (parent_proc_ptr == NULL) {
      perror ("mayGetParentUdp error: unable to malloc");
      return(NULL);
   }
   memset (parent_proc_ptr, 0, SIZEOPRO);
   /* string the message body to the msg head */
   addr_msg.msgptr = (char *) parent_proc_ptr;

   /* go find it.  Make sure we block until the message arrives */
   len = SIZEOPRO;
   may_internal_flag = 234;     /* use mayReceive to get DAEMONMSG */
   flag = mayReceive (NULL, DAEMONMSG, PARENTADDR, 100, &addr_msg, &len);
   may_internal_flag = 0;
   if (flag < 0) { 
      perror ("mayGetParentUdp error: mayrecv failed ");
      return (NULL);
   }

   /* make sure the correct socket fd is recorded. */
   parent_proc_ptr -> socket_id = The_udp_sock -> socket_id;

#ifdef DEBUG
   printf ( " mayGetParentUdp received fd: %d addr: %d 0x%x %d \n", 
   parent_proc_ptr -> socket_id,
   parent_proc_ptr -> address.sin_family,
   parent_proc_ptr -> address.sin_addr.s_addr, 
   parent_proc_ptr -> address.sin_port);
#endif

   return (parent_proc_ptr);

}


/* ============================================================== */
/*  mayGetId 
 * mayGetId  reads a socket descriptor from a file
 * 
 */

#ifdef ANSI_C
void mayGetId (int *sid)
#else
void mayGetId (sid)
int *sid;
#endif
{
   int		pid;		/* the process id of this process */
   char		filename[30];	/* file containing socket id */
   FILE		*fid;		/* file id of above */

   pid = getpid();
   sprintf (filename, "/tmp/may%d", pid);
   fid = fopen (filename, "r");
   if (fid == NULL) {
      *sid = 0;
      return;
   }
   fscanf (fid, "%d", sid);
   fclose (fid);
   if (0 != remove (filename)) {
      perror ("mayGetId: could not remove file");
   }
}

/* ====================== END OF FILE ========================== */
